/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __oiio_ImageOIIO_h__
#define __oiio_ImageOIIO_h__

namespace fe
{

/**************************************************************************//**
    @brief Image handling using OpenImageIO

	@ingroup oiio

	https://sites.google.com/site/openimageio/
*//***************************************************************************/
class FE_DL_EXPORT ImageOIIO: virtual public ImageI
{
	public:
				ImageOIIO(void);
virtual			~ImageOIIO(void);

				//* As ImageI
virtual	U32		createSelect(void);
virtual U32		loadSelect(String filename);
virtual U32		interpretSelect(void* data,U32 size);
virtual U32		interpretSelect(String a_source);
virtual	BWORD	save(String filename);
virtual void	select(U32 id);
virtual	void	unload(U32 id);

virtual	void	setFormat(ImageI::Format a_format);
virtual	ImageI::Format	format(void) const;

virtual	void	resize(U32 width,U32 height,U32 depth);
virtual	void	replaceRegion(U32 x,U32 y,U32 z,
						U32 width,U32 height,U32 depth,void* data);

virtual	U32		width(void) const;
virtual	U32		height(void) const;
virtual	U32		depth(void) const;
virtual	void*	raw(void) const;

virtual	U32		regionCount(void) const					{ return 0; }
virtual	String	regionName(U32 a_regionIndex) const		{ return ""; }
virtual	Box2i	regionBox(String a_regionName) const
				{	return Box2i(Vector2i(0,0),Vector2i(0,0)); }
virtual	String	pickRegion(I32 a_x,I32 a_y) const		{ return ""; }

	private:
		Array<OIIO::ImageBuf*>	m_bufferList;
		I32						m_selected;
};

} // namespace

#endif /* __oiio_ImageOIIO_h__ */
