/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "xml/xml.pmh"

#define FE_SAX_DEBUG	FALSE

namespace fe
{

BWORD SurfaceAccessibleXml::load(String a_filename)
{
#if FE_SAX_DEBUG
	feLog("SurfaceAccessibleXml::load \"%s\"\n",a_filename.c_str());
#endif

	pugi::xml_document document;
	pugi::xml_parse_result result=document.load_file(a_filename.c_str());
	if(!result)
	{
		return result;
	}

	//* TODO param
	const Real triScale=1.0;

	const Color red(1.0,0.0,0.0);
	const Color green(0.0,1.0,0.0);

	sp<SurfaceAccessorI> spOutputPoint=accessor(e_point,e_position,
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputColor=accessor(e_point,"Cd",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputNormal=accessor(e_primitive,"N",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputOrder=accessor(e_primitive,"eulerOrder",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputEuler=accessor(e_primitive,"euler",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputScale=accessor(e_primitive,"scale",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputShear=accessor(e_primitive,"shear",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputName=accessor(e_primitive,"name",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputParent=accessor(e_primitive,"parent",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputRefX=accessor(e_primitive,"refX",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputRefY=accessor(e_primitive,"refY",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputRefZ=accessor(e_primitive,"refZ",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputRefT=accessor(e_primitive,"refT",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputVertices=accessor(e_primitive,e_vertices,
			SurfaceAccessibleI::e_createMissing);

	std::map<String,SpatialTransform> refMap;
	std::map<String,SpatialTransform> animMap;
	std::map<String,SpatialTransform> refTweakMap;
	std::map<String,SpatialTransform> deltaMap;

#if FE_SAX_DEBUG
	feLog("SurfaceAccessibleXml::load result \"%s\"\n",result.description());
#endif

	//* TODO check type and version
	const pugi::xml_node root=document.first_child();

	const pugi::xml_node collection=root.first_child();

	for(pugi::xml_node child=collection.first_child();child;
			child=child.next_sibling())
	{
		const String name=interpretString(child,"name");

		if(name.match(".*_ik_.*") ||
				name.match(".*_goal") ||
				name.match(".*_piv.*") ||
				name.match(".*_pivot") ||
				name.match(".*_skel") ||
				name.match(".*handle") ||
				name.match(".*_pole.*") ||
				name.match(".*_finger.*") ||
				name.match(".*_toe.*"))
		{
			continue;
		}

		const String parent=interpretString(child,"parent");

		const SpatialTransform refWorld=
				interpretTransform(child,"worldreference");
		const SpatialVector worldTranslation=
				interpretVector(child,"translation");
		const SpatialTransform worldRotation=
				interpretEuler(child,"rotation","rotationorder");

		SpatialTransform worldMove;
		setIdentity(worldMove);
		translate(worldMove,worldTranslation);

		SpatialTransform xformParent;
		SpatialTransform refParentTweak;
		SpatialTransform deltaParent;
		if(animMap.find(parent)==animMap.end())
		{
			setIdentity(xformParent);
			setIdentity(refParentTweak);
			setIdentity(deltaParent);
		}
		else
		{
			xformParent=animMap[parent];
			refParentTweak=refTweakMap[parent];
			deltaParent=deltaMap[parent];
		}

		SpatialTransform xformWorld=xformParent;
		xformWorld=worldMove*xformWorld;
		xformWorld=worldRotation*xformWorld;

		refMap[name]=refWorld;
		animMap[name]=xformWorld;

		if(name.match(".*_all.*") ||
				name.match(".*_body") ||
				name.match(".*_space"))

		{
			refTweakMap[name]=refWorld;
			setIdentity(deltaMap[name]);
			continue;
		}

		SpatialTransform invRefWorld;
		invert(invRefWorld,refWorld);
		const SpatialTransform delta=invRefWorld*xformWorld;

		//* find local tweak in reference

		SpatialTransform invParentTweak;
		invert(invParentTweak,refParentTweak);

		SpatialTransform refTweak=refParentTweak;

		const SpatialVector toWorld=
				refWorld.translation()-refParentTweak.translation();
		const SpatialVector toLocal=rotateVector(invParentTweak,toWorld);
		if(magnitudeSquared(toLocal)>1e-6)
		{
			const SpatialVector fromLocal(0.0,1.0,0.0);

			SpatialQuaternion quatLocal;
			set(quatLocal,fromLocal,unitSafe(toLocal));

			const SpatialTransform localTweak=quatLocal;

			refTweak=localTweak*refTweak;
		}
		setTranslation(refTweak,refWorld.translation());

		SpatialTransform animTweak=refTweak*deltaParent;
		setTranslation(animTweak,xformWorld.translation());

		refTweakMap[name]=refTweak;
		deltaMap[name]=delta;

		const String order=interpretString(child,"rotationorder");
		const SpatialVector degrees=interpretVector(child,"rotation");
		const SpatialVector scale=interpretVector(child,"scale");
		const SpatialVector shear=interpretVector(child,"shear");
		const SpatialVector color=interpretVector(child,"color");

		const I32 primitiveIndex=spOutputVertices->append(3);
		spOutputName->set(primitiveIndex,name);
		spOutputParent->set(primitiveIndex,parent);
		spOutputNormal->set(primitiveIndex,animTweak.column(1));
		spOutputOrder->set(primitiveIndex,order);
		spOutputEuler->set(primitiveIndex,degrees);
		spOutputScale->set(primitiveIndex,scale);
		spOutputShear->set(primitiveIndex,shear);

		spOutputRefX->set(primitiveIndex,refTweak.column(0));
		spOutputRefY->set(primitiveIndex,refTweak.column(1));
		spOutputRefZ->set(primitiveIndex,refTweak.column(2));
		spOutputRefT->set(primitiveIndex,refTweak.translation());

		const I32 pointIndex0=spOutputVertices->integer(primitiveIndex,0);
		const I32 pointIndex1=spOutputVertices->integer(primitiveIndex,1);
		const I32 pointIndex2=spOutputVertices->integer(primitiveIndex,2);

		const SpatialVector point=animTweak.translation();
		spOutputPoint->set(pointIndex0,point);
		spOutputPoint->set(pointIndex1,point+triScale*animTweak.column(0));
		spOutputPoint->set(pointIndex2,point+triScale*animTweak.column(2));

		spOutputColor->set(pointIndex0,color);
		spOutputColor->set(pointIndex1,red);
		spOutputColor->set(pointIndex2,green);

#if FE_SAX_DEBUG

		feLog("%s \"%s\" parent \"%s\" order \"%s\"\n",
				child.name(),name.c_str(),parent.c_str(),order.c_str());
		feLog("  T %s R %s S %s sh %s\n",
				c_print(worldTranslation),c_print(degrees),
				c_print(scale),c_print(shear));
		feLog("  rot\n%s\n",c_print(worldRotation));
		feLog("  xformWorld\n%s\n",c_print(xformWorld));
		feLog("  ref\n%s\n",c_print(refWorld));
		feLog("  delta\n%s\n",c_print(refWorld));
		feLog("  refTweak\n%s\n",c_print(refTweak));
		feLog("  animTweak\n%s\n",c_print(animTweak));
		feLog("  prim %d points %d %d %d\n\n",
				primitiveIndex,pointIndex0,pointIndex1,pointIndex2);
#endif
	}

	return result;
}

SpatialTransform SurfaceAccessibleXml::interpretEuler(
	pugi::xml_node& a_rNode,String a_attrName,String a_attrOrder)
{
	const SpatialVector degrees=interpretVector(a_rNode,a_attrName);
	const String order=interpretString(a_rNode,a_attrOrder);

	const SpatialEuler euler(
			degrees[0]*degToRad,degrees[1]*degToRad,degrees[2]*degToRad);

	if(order=="ZYX")
	{
		return euler;
	}

	//* TODO move to Euler<>

	SpatialTransform xform;
	setIdentity(xform);

	if(order=="ZXY")
	{
		rotate(xform,euler[2],e_zAxis);
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[1],e_yAxis);
	}
	else if(order=="XZY")
	{
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[2],e_zAxis);
		rotate(xform,euler[1],e_yAxis);
	}
	else if(order=="YZX")
	{
		rotate(xform,euler[1],e_yAxis);
		rotate(xform,euler[2],e_zAxis);
		rotate(xform,euler[0],e_xAxis);
	}
	else if(order=="XYZ")
	{
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[1],e_yAxis);
		rotate(xform,euler[2],e_zAxis);
	}
	else if(order=="YXZ")
	{
		rotate(xform,euler[1],e_yAxis);
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[2],e_zAxis);
	}

	return xform;
}

String SurfaceAccessibleXml::interpretString(
	pugi::xml_node& a_rNode,String a_attrName)
{
	pugi::xml_attribute xmlAttr=a_rNode.attribute(a_attrName.c_str());

	return xmlAttr.value();
}

SpatialVector SurfaceAccessibleXml::interpretVector(
	pugi::xml_node& a_rNode,String a_attrName)
{
	pugi::xml_attribute xmlAttr=a_rNode.attribute(a_attrName.c_str());

	float x=0.0;
	float y=0.0;
	float z=0.0;
	sscanf(xmlAttr.value(),"%f,%f,%f",&x,&y,&z);

	return SpatialVector(x,y,z);
}

SpatialTransform SurfaceAccessibleXml::interpretTransform(
	pugi::xml_node& a_rNode,String a_attrName)
{
	pugi::xml_attribute xmlAttr=a_rNode.attribute(a_attrName.c_str());

	float v[16];
	sscanf(xmlAttr.value(),
			"%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f",
			&v[0],&v[1],&v[2],&v[3],
			&v[4],&v[5],&v[6],&v[7],
			&v[8],&v[9],&v[10],&v[11],
			&v[12],&v[13],&v[14],&v[15]);

	SpatialTransform matrix;
	set(matrix,v);

	return matrix;
}

} // namespace
